def serialize_result(
    message_key, serializable_result, start_time, duration_seconds, runtime_info=None
):
    """
    Wrap the computation result in the `werkit.Manager` result schema.

    Args:
        serializable_result (object): Any serializable result.
        start_time (datetime.datetime): The start time of the computation.
        duration_seconds (int): The computation time.
        runtime_info (object): Serializable runtime metadata.
    """
    return {
        "message_key": message_key,
        "success": True,
        "result": serializable_result,
        "error": None,
        "error_origin": None,
        "start_time": start_time.astimezone().isoformat(),
        "duration_seconds": duration_seconds,
        "runtime_info": runtime_info,
    }


def serialize_exception(
    message_key,
    exception,
    error_origin,
    start_time,
    duration_seconds=-1,
    runtime_info=None,
):
    """
    Wrap an exception in the `werkit.Manager` result schema.

    Args:
        exception (Exception): The exception to wrap.
        error_origin (str): Either `"compute"`, `"system"` or
            `"orchestration"`. Refer to the `werkit.Manager` documentation for
            the definitions.
        start_time (datetime.datetime): The start time of the computation.
        duration_seconds (int): The time elapsed before the error occurred,
            if known.
        runtime_info (object): Serializable runtime metadata.
    """
    import traceback

    return {
        "message_key": message_key,
        "success": False,
        "result": None,
        "error": traceback.format_exception(None, exception, exception.__traceback__),
        "error_origin": error_origin,
        "duration_seconds": duration_seconds,
        "start_time": start_time.astimezone().isoformat(),
        "runtime_info": runtime_info,
    }



import os
import shutil
import sys
import venv
import zipfile
from executor import execute


def export_poetry_requirements(
    output_file, extras=[], with_credentials=True, with_hashes=True
):
    args = ["poetry", "export", "--output", output_file]
    if with_credentials:
        args.append("--with-credentials")
    if not with_hashes:
        args.append("--without-hashes")
    for extra in extras:
        args += ["--extras", extra]
    execute(*args)


def create_venv_with_dependencies(
    venv_dir,
    upgrade_pip=True,
    install_wheel=True,
    install_requirements_from=["requirements.txt"],
    install_transitive_dependencies=True,
    environment={},
):
    venv.create(venv_dir, with_pip=True)
    python = os.path.join(venv_dir, "bin", "python")

    if upgrade_pip:
        execute(
            python, "-m", "pip", "install", "--upgrade", "pip", environment=environment
        )

    if install_wheel:
        execute(python, "-m", "pip", "install", "wheel", environment=environment)

    if len(install_requirements_from) > 0:
        args = [python, "-m", "pip", "install"]
        if not install_transitive_dependencies:
            args += ["--no-dependencies"]
        for requirements_file in install_requirements_from:
            args += ["-r", requirements_file]
        execute(*args, environment=environment)


def find_site_packages_dir(venv_dir):
    for candidate_result in [
        os.path.join(venv_dir, "lib64", "python3.8", "site-packages"),
        os.path.join(venv_dir, "lib", "python3.8", "site-packages"),
        os.path.join(venv_dir, "lib64", "python3.7", "site-packages"),
        os.path.join(venv_dir, "lib", "python3.7", "site-packages"),
    ]:
        if os.path.exists(candidate_result):
            return candidate_result
    raise ValueError(f"Unable to locate site-packages folder in {venv_dir}")


def collect_zipfile_contents(
    target_dir, venv_dir, src_files=[], src_dirs=[], lib_files=[], verbose=False
):
    def pif(x):
        if verbose:
            print(x, file=sys.stderr)

    if os.path.isdir(target_dir):
        raise ValueError(f"target_dir should not exist: {target_dir}")
    if not os.path.isdir(venv_dir):
        raise ValueError(f"venv_dir should already be populated: {venv_dir}")

    # Copy dependencies from venv.
    site_packages_dir = find_site_packages_dir(venv_dir)
    pif(f"Copying dependencies from {site_packages_dir} to {target_dir}")
    shutil.copytree(site_packages_dir, target_dir)

    if lib_files:
        lib_dir = os.path.join(target_dir, "lib")
        os.makedirs(lib_dir, exist_ok=True)
        for lib_file in lib_files:
            target = os.path.join(lib_dir, os.path.basename(lib_file))
            pif(f"Copying {lib_file} to {target}/")
            shutil.copyfile(lib_file, target)

    for src_dir in src_dirs:
        target = os.path.join(target_dir, src_dir)
        pif(f"Copying {src_dir} to {target}")
        shutil.copytree(src_dir, target)

    for src_file in src_files:
        target = os.path.join(target_dir, src_file)
        pif(f"Copying {src_file} to {target}")
        shutil.copyfile(src_file, target)


def create_zipfile_from_dir(dir_path, path_to_zipfile):
    with zipfile.ZipFile(path_to_zipfile, "w", zipfile.ZIP_DEFLATED) as f:
        for root, dirs, files in os.walk(dir_path):
            for file in files:
                fullpath = os.path.join(root, file)
                arcname = os.path.relpath(fullpath, start=dir_path)
                f.write(fullpath, arcname=arcname)



import os
import sys
import uuid
from contextlib import contextmanager


@contextmanager
def temp_file_on_s3(local_path, bucket, key=None, verbose=False):
    """
    Copy the given path to S3. Delete the object from S3 when the block
    exits.
    """
    import boto3

    def pif(x):
        if verbose:
            print(x, file=sys.stderr)

    if key is None:
        filename, extension = os.path.splitext(os.path.basename(local_path))
        key = f"{filename}_{uuid.uuid4().hex}{extension}"

    s3_client = boto3.client("s3")

    file_on_s3 = f"s3://{bucket}/{key}"
    pif(f"Uploading {local_path} to {file_on_s3}")
    s3_client.upload_file(Filename=local_path, Bucket=bucket, Key=key)

    try:
        yield key
    finally:
        pif(f"Removing {file_on_s3}")
        s3_client.delete_object(Bucket=bucket, Key=key)


@contextmanager
def temp_file_on_s3_from_string(
    contents, bucket, key=None, extension=None, verbose=False, ret_etag=False
):
    """
    Write the given contents to S3. Delete the object from S3 when the block
    exits.
    """
    import boto3

    def pif(x):
        if verbose:
            print(x, file=sys.stderr)

    if key is None:
        key = f"{uuid.uuid4().hex}{extension}"

    s3_client = boto3.client("s3")

    file_on_s3 = f"s3://{bucket}/{key}"
    pif(f"Writing to {file_on_s3}")

    response = s3_client.put_object(
        Bucket=bucket, Key=key, Body=contents.encode("utf-8")
    )

    try:
        yield (key, response["ETag"]) if ret_etag else key
    finally:
        pif(f"Removing {file_on_s3}")
        s3_client.delete_object(Bucket=bucket, Key=key)



import typing as t
from abc import ABC, abstractmethod

JSONType = t.Union[str, int, float, bool, None, t.Dict[str, t.Any], t.List[t.Any]]

CanonicalType = t.TypeVar("CanonicalType")


class CustomType(ABC, t.Generic[CanonicalType]):
    """
    While the canonical type could be the type class itself, this isn't required
    or even encouraged. Any existing type can be used as the canonical type,
    whereas the CustomType is responsible for canonicalization and
    serialization.
    """

    @classmethod
    def name(cls) -> str:
        """
        The name of the custom type, used in the serialized dependency graph.
        The default is `cls.__name__`.
        """
        return cls.__name__

    @classmethod
    def schema_path(cls) -> str:
        """
        The path to the JSON schema. The default is 'generated/schema.json',
        relative to the file where the class is defined.
        """
        import inspect
        import os

        return os.path.join(
            os.path.dirname(inspect.getfile(cls)), "generated", "schema.json"
        )

    @classmethod
    def ref(cls) -> str:
        """
        The ref to look up in the JSON schema. The default is
        '#/definitions/{cls.name()}'.
        """
        return f"#/definitions/{cls.name()}"

    @classmethod
    def validate(cls, json_data: JSONType) -> None:
        """
        Validate the JSON representation.
        """
        import simplejson as json
        from jsonschema import Draft7Validator, RefResolver

        try:
            validator = cls._validator
        except AttributeError:
            with open(cls.schema_path(), "r") as f:
                schema = json.load(f)

            resolver = RefResolver.from_schema(schema)
            validator = cls._validator = Draft7Validator(
                {"$ref": cls.ref()}, resolver=resolver
            )

        validator.validate(json_data)

    @classmethod
    @abstractmethod
    def deserialize(cls, json_data: JSONType) -> CanonicalType:
        """
        Convert the JSON representation to the canonical native type.
        """

    @classmethod
    @abstractmethod
    def coerce(cls, value: t.Any) -> CanonicalType:
        """
        Coerce the given value to the canonical native type. Raise an exception
        if it can't be coerced.
        """

    @classmethod
    @abstractmethod
    def serialize(self, value: CanonicalType) -> JSONType:
        """
        Convert the canonical native type to a JSON representation.
        """



import typing as t
from ._dependency_graph import DependencyGraph


class StateManager:
    def __init__(self, instance: t.Any):
        self.instance = instance
        self.dependency_graph = DependencyGraph(instance.__class__)
        self.store: t.Dict = {}

    def _assert_known_keys(self, keys: t.Iterable[str]) -> None:
        unknown_keys = set(keys) - set(self.dependency_graph.keys())
        if len(unknown_keys):
            raise KeyError(
                f"Unknown {'key' if len(unknown_keys) == 1 else 'keys'}: {', '.join(sorted(list(unknown_keys)))}"
            )

    def deserialize(self, **kwargs: t.Dict) -> None:
        self._assert_known_keys(kwargs.keys())
        deserialized = {
            name: self.dependency_graph.all_nodes[name].deserialize(value)
            for name, value in kwargs.items()
        }
        self.store.update(deserialized)

    def coerce(self, **kwargs: t.Dict) -> t.Dict:
        return {
            name: self.dependency_graph.all_nodes[name].coerce(name=name, value=value)
            for name, value in kwargs.items()
        }

    def set(self, **kwargs: t.Dict) -> None:
        self._assert_known_keys(kwargs.keys())
        coerced = self.coerce(**kwargs)
        self.store.update(coerced)

    def evaluate(
        self, targets: t.List[str] = None, handle_exceptions: bool = False
    ) -> None:
        from artifax import Artifax

        if targets is not None:
            if len(targets) == 0:
                # TODO: afx.build(targets=[]) raises an exception. This seems
                # like a bug in Artifax.
                return
            else:
                self._assert_known_keys(targets)

        afx = Artifax(
            {
                name: self.dependency_graph.compute_nodes[name].bind(self.instance)
                for name in self.dependency_graph.compute_nodes.keys()
            }
        )
        if self.store:
            afx.set(**self.store)
        afx.build(targets=targets)
        # TODO: `afx.build()` should always return an object.
        coerced = self.coerce(**afx._result)
        self.store.update(coerced)

    def serialize(self, targets: t.List[str] = None) -> t.Dict:
        self.evaluate(targets=targets)

        return {
            name: self.dependency_graph.all_nodes[name].serialize_value(value)
            for name, value in self.store.items()
            if targets is None or name in targets
        }

    def get(self, name: str) -> t.Any:
        self._assert_known_keys([name])

        try:
            return self.store[name]
        except KeyError:
            pass

        if name in self.dependency_graph.inputs:
            raise KeyError(f"Input has not been set: {name}")

        if name in self.dependency_graph.compute_nodes.keys():
            self.evaluate(targets=[name])

        return self.store[name]



import os
import sys
import boto3
from ..s3 import temp_file_on_s3

DEFAULT_RUNTIME = "python3.7"


def needs_s3_upload(path_to_zipfile):
    return os.path.getsize(path_to_zipfile) > 50 * 2**20


def _create_or_update(
    local_path_to_zipfile,
    s3_path_to_zipfile,
    message,
    create_or_update_fn,
    verbose=False,
    s3_code_bucket=None,
    force_upload_to_s3_code_bucket=False,
):
    """
    Create or update a lambda function with the given zipfile. If the zipfile is larger
    than 50 MB, you must specify an `s3_code_bucket`.
    """
    if s3_path_to_zipfile and not s3_code_bucket:
        raise ValueError(
            "When `s3_path_to_zipfile` is provided, `s3_code_bucket` is required"
        )

    should_upload_directly_to_lambda = (
        local_path_to_zipfile
        and not force_upload_to_s3_code_bucket
        and not needs_s3_upload(local_path_to_zipfile)
    )

    if not should_upload_directly_to_lambda and not s3_code_bucket:
        raise ValueError(
            """
            When zipfile is larger than 50 MB,
            or force_upload_to_s3_code_bucket is True,
            then s3_code_bucket is required
            """
        )

    if local_path_to_zipfile and not os.path.isfile(local_path_to_zipfile):
        raise ValueError(f"Local zip file does not exist: {local_path_to_zipfile}")

    def pif(x):
        if verbose:
            print(x, file=sys.stderr)

    if should_upload_directly_to_lambda:
        with open(local_path_to_zipfile, "rb") as f:
            zipfile_contents = f.read()
        pif(f"Uploading {local_path_to_zipfile} to Lambda")
        create_or_update_fn({"ZipFile": zipfile_contents})
        pif(message)
    elif s3_path_to_zipfile:
        create_or_update_fn({"S3Bucket": s3_code_bucket, "S3Key": s3_path_to_zipfile})
        pif(message)
    else:
        with temp_file_on_s3(
            local_path=local_path_to_zipfile, bucket=s3_code_bucket, verbose=verbose
        ) as temp_key:
            create_or_update_fn({"S3Bucket": s3_code_bucket, "S3Key": temp_key})
            pif(message)


def perform_create(
    aws_region,
    handler,
    function_name,
    role,
    local_path_to_zipfile=None,
    s3_path_to_zipfile=None,
    timeout=None,
    memory_size=None,
    runtime=DEFAULT_RUNTIME,
    env_vars={},
    s3_code_bucket=None,
    verbose=False,
    force_upload_to_s3_code_bucket=False,
    wait_until_active=True,
    wait_until_active_timeout_seconds=30,
):
    if bool(local_path_to_zipfile) == bool(s3_path_to_zipfile):
        raise ValueError(
            "Either `local_path_to_zipfile` or `s3_path_to_zipfile` must be provided (and not both)"
        )

    client = boto3.client("lambda", region_name=aws_region)

    def create(code_arguments):
        extra_options = {}
        if timeout is not None:
            extra_options["Timeout"] = timeout
        if memory_size is not None:
            extra_options["MemorySize"] = memory_size

        client.create_function(
            FunctionName=function_name,
            Runtime=runtime,
            Role=role,
            Handler=handler,
            Environment={"Variables": env_vars},
            Code=code_arguments,
            **extra_options,
        )

    _create_or_update(
        local_path_to_zipfile=local_path_to_zipfile,
        s3_path_to_zipfile=s3_path_to_zipfile,
        message="Lambda function created",
        create_or_update_fn=create,
        verbose=verbose,
        s3_code_bucket=s3_code_bucket,
        force_upload_to_s3_code_bucket=force_upload_to_s3_code_bucket,
    )

    if wait_until_active:
        client.get_waiter("function_active").wait(
            FunctionName=function_name,
            WaiterConfig={
                "Delay": 1,
                "MaxAttempts": wait_until_active_timeout_seconds,
            },
        )


def perform_update_code(
    aws_region,
    function_name,
    local_path_to_zipfile=None,
    s3_path_to_zipfile=None,
    s3_code_bucket=None,
    verbose=False,
    force_upload_to_s3_code_bucket=False,
    wait_until_updated=True,
    wait_until_updated_timeout_seconds=30,
):
    if bool(local_path_to_zipfile) == bool(s3_path_to_zipfile):
        raise ValueError(
            "Either `local_path_to_zipfile` or `s3_path_to_zipfile` must be provided (and not both)"
        )

    client = boto3.client("lambda", region_name=aws_region)

    def update(code_arguments):
        client.update_function_code(FunctionName=function_name, **code_arguments)

    _create_or_update(
        local_path_to_zipfile=local_path_to_zipfile,
        s3_path_to_zipfile=s3_path_to_zipfile,
        message="Lambda function code updated",
        create_or_update_fn=update,
        verbose=verbose,
        s3_code_bucket=s3_code_bucket,
        force_upload_to_s3_code_bucket=force_upload_to_s3_code_bucket,
    )

    if wait_until_updated:
        client.get_waiter("function_updated").wait(
            FunctionName=function_name,
            WaiterConfig={
                "Delay": 1,
                "MaxAttempts": wait_until_updated_timeout_seconds,
            },
        )



from __future__ import annotations
import inspect
import numbers
import typing as t
from typing_extensions import Literal, TypedDict
from ._built_in_type import (
    BuiltInValueType,
    coerce_value_to_builtin_type,
    is_built_in_value_type,
)
from ._custom_type import CustomType, JSONType


AnyValueType = t.Union[BuiltInValueType, t.Type[CustomType]]


class BaseNode:
    def __init__(self, value_type: AnyValueType):
        if is_built_in_value_type(value_type) or (
            inspect.isclass(value_type) and issubclass(value_type, CustomType)
        ):
            self.value_type = value_type
        else:
            raise ValueError(
                "Expected value type to be bool, int, float, str, or a subclass of CustomType"
            )

    @property
    def value_type_is_built_in(self) -> bool:
        return is_built_in_value_type(self.value_type)

    @property
    def value_type_name(self) -> str:
        if self.value_type_is_built_in:
            if self.value_type is bool:
                return "boolean"
            elif self.value_type in (int, float, numbers.Number):
                return "number"
            elif self.value_type is str:
                return "string"
            else:
                raise ValueError("How did we get here?")
        else:
            return self.value_type.__name__

    def deserialize(self, value: JSONType) -> t.Any:
        if self.value_type_is_built_in:
            return value
        else:
            value_type = t.cast(t.Type[CustomType], self.value_type)
            value_type.validate(value)
            return value_type.deserialize(value)

    def coerce(self, name: str, value: t.Any) -> t.Any:
        if self.value_type_is_built_in:
            return coerce_value_to_builtin_type(
                name=name,
                value_type=t.cast(BuiltInValueType, self.value_type),
                value=value,
            )
        else:
            # TODO: Perhaps catch and re-throw to improve the error message.
            return t.cast(t.Type[CustomType], self.value_type).coerce(value)

    def serialize_value(self, value: t.Any) -> JSONType:
        if self.value_type_is_built_in:
            return value
        else:
            value_type = t.cast(t.Type[CustomType], self.value_type)
            serialized = value_type.serialize(value)
            value_type.validate(serialized)
            return serialized


InputJSONType = TypedDict("InputJSONType", {"valueType": str})


class Input(BaseNode):
    def serialize(self) -> InputJSONType:
        return {"valueType": self.value_type_name}


ComputeNodeJSONType = TypedDict(
    "ComputeNodeJSONType", {"valueType": str, "dependencies": t.List[str]}
)


class ComputeNode(BaseNode):
    # TODO: Use a narrower type for `method`.
    def __init__(self, method: t.Callable, value_type: AnyValueType):
        super().__init__(value_type=value_type)

        self.method = method
        self.dependencies = [x for x in inspect.signature(method).parameters.keys()][1:]

    def bind(self, instance: ComputeNode) -> t.Callable:
        import functools

        return functools.partial(self.method, instance)

    def serialize(self) -> ComputeNodeJSONType:
        return {
            "valueType": self.value_type_name,
            "dependencies": self.dependencies,
        }


class Intermediate(ComputeNode):
    pass


class Output(ComputeNode):
    pass


def intermediate(value_type: AnyValueType):
    def decorator(method):
        return Intermediate(method, value_type)

    return decorator


def output(value_type: AnyValueType):
    def decorator(method):
        return Output(method, value_type)

    return decorator


DependencyGraphJSONType = TypedDict(
    "DependencyGraphJSONType",
    {
        "schemaVersion": Literal[1],
        "inputs": t.Dict[str, InputJSONType],
        "intermediates": t.Dict[str, ComputeNodeJSONType],
        "outputs": t.Dict[str, ComputeNodeJSONType],
    },
)

AttrType = t.TypeVar("AttrType")


def _attrs_of_type(obj: t.Any, _type: t.Type[AttrType]) -> t.Dict[str, AttrType]:
    return {
        name: getattr(obj, name)
        for name in dir(obj)
        if not name.startswith("__") and isinstance(getattr(obj, name), _type)
    }


class DependencyGraph:
    def __init__(self, cls: t.Type):
        assert inspect.isclass(cls)

        self.inputs = _attrs_of_type(cls, Input)
        self.intermediates = _attrs_of_type(cls, Intermediate)
        self.outputs = _attrs_of_type(cls, Output)
        self.compute_nodes = dict(**self.intermediates, **self.outputs)
        self.all_nodes = dict(**self.inputs, **self.compute_nodes)

    def keys(self) -> t.List[str]:
        return list(self.inputs.keys()) + list(self.compute_nodes.keys())

    def serialize(self) -> DependencyGraphJSONType:
        return {
            "schemaVersion": 1,
            "inputs": {k: v.serialize() for k, v in self.inputs.items()},
            "intermediates": {k: v.serialize() for k, v in self.intermediates.items()},
            "outputs": {k: v.serialize() for k, v in self.outputs.items()},
        }



import typing as t
from ._dependency_graph import ComputeNode, Input
from ._state_manager import StateManager


def bind_state_manager(attr_name="state_manager"):
    def decorator(cls):
        class Bound(cls):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)
                setattr(self, attr_name, StateManager(self))

            def __getattribute__(self, name: str) -> t.Any:
                attr = object.__getattribute__(self, name)
                if isinstance(attr, Input) or isinstance(attr, ComputeNode):
                    state_manager = object.__getattribute__(self, attr_name)
                    return state_manager.get(name)
                else:
                    return attr

        return Bound

    return decorator



import datetime
import sys
from ._formatting import format_time
from ._serialization import serialize_exception, serialize_result


class Manager:
    """
    Wrap a bit of cloud or batch computation, automatically handling and
    serializing errors and timing the computation.

    Either set `manager.result` with the result of the computation or raise an
    exception. When the handler closes, `manager.output_message` will be a
    dictionary with the following keys:

    - `success` (`bool`): `True` so long as no exception was raised.
    - `result` (`object`): In case of succes, the deserialized result.
    - `error` (`None` or `list`): In case of an error, the formatted exception
        traceback.
    - `error_origin` (`None` or `str`): In case of an error, the source of the
        error, which is `'compute'` for errors which originate inside the
        context handler. Other values are reserved (but unsued by this
        function): `'system'` for errors on the compute system which are
        external to the compute code, and `'orchestration'` for other errors
        which occur while attempting to reach the remote system, including
        timeouts.
    - `start_time` (`str`): The local start time of the computation, in ISO
        8601 with the time zone.
    - `duration_seconds` (`float`): The time spent in computation.
    - `runtime_info` (`object`): User-provided runtime metadata.

    Args:
        schema (werkit.compute.Schema): A helper object containing the
            request, result, and serialized result schemas.
        destination (werkit.compute.Destination): An optional destination
            to which the serialized result is dispatched.
        handle_exceptions (bool): When `True`, exceptions thrown within the
            block are caught and serialized, as is useful for wrapping cloud
            processes or batch jobs, where you don't want to crash. When
            `False` these exceptions propagate, as is useful for development.
            `KeyboardInterrupt` and `SystemExit` exceptions always propagate.
        vebose (bool): When `True`, print out timing information when
            computation finishes.
        time_precision (int): The number of decimal places (of seconds) to
            include in the timing result. The default is 2, which rounds to
            the nearest hundredth of a second.

    Example:

        with Manager(
            input_message=params,
            schema=schema,
            handle_exceptions=handle_exceptions,
            verbose=verbose,
        ) as manager:
            result = perform_computation()
            # If needed, convert result to JSON-serializable objects.
            serialized = transform(result)
            manager.result = serialized
        return manager.output_message
    """

    def __init__(
        self,
        input_message,
        schema,
        destination=None,
        runtime_info=None,
        handle_exceptions=True,
        verbose=False,
        time_precision=2,
    ):
        # Import late to avoid circular import.
        from werkit.compute import Destination, Schema

        if not isinstance(schema, Schema):
            raise ValueError(
                "Expected schema to be an instance of werkit.compute.Schema"
            )
        self.schema = schema

        if destination is not None and not isinstance(destination, Destination):
            raise ValueError(
                "Expected destination to be an instance of werkit.compute.Destination"
            )
        self.destination = destination

        self.runtime_info = runtime_info
        self.handle_exceptions = handle_exceptions
        self.verbose = verbose
        self.time_precision = time_precision
        self.input_message = input_message

    @property
    def message_key(self):
        try:
            return self.input_message["message_key"]
        except (KeyError, TypeError):
            raise ValueError("Input message is missing `message_key` property")

    @property
    def duration_seconds(self):
        return round(
            (datetime.datetime.now() - self.start_time).total_seconds(),
            self.time_precision,
        )

    def __enter__(self):
        self.start_time = datetime.datetime.now()
        try:
            self.schema.input_message.validate(self.input_message)
        except:  # noqa: E722
            if not self.__exit__(*sys.exc_info()):
                raise
        return self

    @property
    def result(self):
        return self._result

    @result.setter
    def result(self, value):
        self.schema.output.validate(value)
        self._result = value

    def serialize_result(self, result):
        return serialize_result(
            message_key=self.message_key,
            serializable_result=result,
            start_time=self.start_time,
            duration_seconds=self.duration_seconds,
            runtime_info=self.runtime_info,
        )

    def _note_compute_success(self, result):
        self.output_message = self.serialize_result(result)
        self.schema.output_message.validate(self.output_message)
        if self.destination:
            self.destination.send(
                message_key=self.message_key, output_message=self.output_message
            )

    def serialize_exception(self, exception):
        return serialize_exception(
            message_key=self.message_key,
            exception=exception,
            error_origin="compute",
            start_time=self.start_time,
            duration_seconds=self.duration_seconds,
            runtime_info=self.runtime_info,
        )

    def _note_compute_exception(self, exception):
        """
        Return a value suitable for returning from `__exit__`.
        """
        self.output_message = self.serialize_exception(exception)
        self.schema.output_message.validate(self.output_message)
        if self.handle_exceptions:
            print(
                "Error handled by werkit. (To disable, invoke `Manager()` with `handle_exceptions=False`.)"
            )
            print("".join(self.output_message["error"]))
            if self.destination:
                self.destination.send(
                    message_key=self.message_key, output_message=self.output_message
                )
            return True
        else:
            return False

    def __exit__(self, type, value, traceback):
        if type in [KeyboardInterrupt, SystemExit]:
            raise value
        elif value:
            if self.verbose:
                print(
                    "Errored in {}".format(format_time(self.duration_seconds)),
                    file=sys.stderr,
                )
            return self._note_compute_exception(value)

        # In case of success, make sure the `result` setter has been invoked
        # inside the block.
        try:
            self.result
        except AttributeError:
            return self._note_compute_exception(
                AttributeError("'result' was not set on the 'Manager' instance")
            )

        self._note_compute_success(self.result)

        if self.verbose:
            print(
                "Completed in {}".format(format_time(self.duration_seconds)),
                file=sys.stderr,
            )



import typing as t
from . import CustomType, Input, JSONType, bind_state_manager, intermediate, output


@bind_state_manager()
class MyComputeProcess:
    a = Input(value_type=int)
    b = Input(value_type=int)

    @intermediate(value_type=int)
    def i(self, a: int) -> int:
        return a

    @intermediate(value_type=int)
    def j(self, b: int) -> int:
        return b

    @output(value_type=int)
    def r(self, i: int, j: int) -> int:
        return i + j


EXPECTED_INPUTS = ["a", "b"]
EXPECTED_INTERMEDIATES = ["i", "j"]
EXPECTED_OUTPUT = "r"

EXPECTED_SERIALIZED_DEPENDENCY_GRAPH = {
    "schemaVersion": 1,
    "inputs": {"a": {"valueType": "number"}, "b": {"valueType": "number"}},
    "intermediates": {
        "i": {"valueType": "number", "dependencies": ["a"]},
        "j": {"valueType": "number", "dependencies": ["b"]},
    },
    "outputs": {"r": {"valueType": "number", "dependencies": ["i", "j"]}},
}


class MyComputeProcessSubclass(MyComputeProcess):
    pass


class MyRaisingComputeProcess(MyComputeProcess):
    @output(value_type=int)
    def s(self) -> int:
        raise ValueError("Whoops")

    @output(value_type=int)
    def t(self) -> int:
        raise ValueError("Whoops")


class MyWronglyTypedComputeProcess(MyComputeProcess):
    @output(value_type=int)
    def s(self) -> bool:
        return False


class MyModel:
    def __init__(self, title: str, description: str, count: int):
        if title is None or description is None or count is None:
            raise ValueError("All properties are required")
        self.title = title
        self.description = description
        self.count = count


class MyModelType(CustomType[MyModel]):
    @classmethod
    def name(cls) -> str:
        return "MyModel"

    @classmethod
    def coerce(cls, value: t.Any) -> MyModel:
        if not isinstance(value, MyModel):
            raise ValueError(f"Can't coerce {type(value).__name__} to {cls.__name__}")
        return value

    @classmethod
    def serialize(self, value: MyModel) -> JSONType:
        return {
            "title": value.title,
            "description": value.description,
            "count": value.count,
        }

    @classmethod
    def deserialize(self, json_data: JSONType) -> MyModel:
        json_data = t.cast(dict, json_data)
        return MyModel(**json_data)


class Vector3(CustomType[tuple]):
    DECIMALS = 2

    @classmethod
    def coerce(cls, value: t.Any) -> tuple:
        if not isinstance(value, tuple):
            raise ValueError(f"Can't coerce {type(value).__name__} to {cls.__name__}")
        elif not len(value) == 3:
            raise ValueError("Excepted tuple to have length 3")
        return tuple(round(coord, cls.DECIMALS) for coord in value)

    @classmethod
    def serialize(self, value: tuple) -> JSONType:
        return list(value)

    @classmethod
    def deserialize(self, json_data: JSONType) -> tuple:
        json_data = t.cast(list, json_data)
        return tuple(json_data)


class MyComputeProcessWithCustomType(MyComputeProcess):
    @output(value_type=MyModelType)
    def thing(self) -> MyModel:
        return MyModel(
            title="Example title", description="Example description", count=25
        )

    @output(value_type=Vector3)
    def other_thing(self) -> tuple:
        return (1.5151, 2.5151, 3.5151)



import asyncio
import datetime
import json
import boto3
from botocore.exceptions import ClientError
from harrison import Timer


lambda_client = boto3.client("lambda")
event_loop = asyncio.get_event_loop()


async def call_worker_service(
    worker_lambda_function_name,
    _input,
    event_loop=event_loop,
    executor=None,
):
    def invoke_lambda():
        return lambda_client.invoke(
            FunctionName=worker_lambda_function_name, Payload=json.dumps(_input)
        )

    start_timestamp = datetime.datetime.utcnow().timestamp()
    with Timer(verbose=False) as roundtrip_timer:
        response = await event_loop.run_in_executor(executor, invoke_lambda)
        payload = await event_loop.run_in_executor(executor, response["Payload"].read)
    output = json.loads(payload.decode())
    output["orchestrationStartTimestamp"] = start_timestamp
    output["workerRoundtripSeconds"] = roundtrip_timer.elapsed_time_s
    return output


# TODO: This is imported from some utility code which invokes a lambda.
# Generalize this function and move it outside this `orchestrator_lambda`
# module.
async def wait_for(
    _input,
    timeout,
    worker_lambda_function_name,
    event_loop=event_loop,
    executor=None,
):
    try:
        return await asyncio.wait_for(
            call_worker_service(
                worker_lambda_function_name=worker_lambda_function_name,
                _input=_input,
                event_loop=event_loop,
                executor=executor,
            ),
            timeout=timeout,
        )
    except ClientError as ex:
        return ex
    except asyncio.TimeoutError as ex:
        return ex
    except Exception as ex:
        return ex


async def parallel_map_on_lambda(
    message_key,
    item_property_name,
    item_collection,
    common_input,
    worker_lambda_function_name,
    timeout,
    event_loop=event_loop,
    executor=None,
):
    coroutines = [
        wait_for(
            _input={
                **common_input,
                "message_key": message_key,
                "itemKey": item_key,
                item_property_name: item,
            },
            worker_lambda_function_name=worker_lambda_function_name,
            timeout=timeout,
            event_loop=event_loop,
            executor=executor,
        )
        for item_key, item in item_collection.items()
    ]
    return await asyncio.gather(*coroutines, return_exceptions=True)
